"""basic tests for serialization of objects"""

import unittest

from hein_control.mixin.serialization import (DumpableState, DumpableList, is_jsonable, make_mapping_serializable,
                                              make_iterable_serializable)

class State(DumpableState):
    val = 1234

    def as_dict(self) -> dict:
        return {'test value': self.val}


class List(DumpableList):
    lst = [1, 2, 3, 4, 5, 6, 'asdf']

    def as_list(self) -> list:
        return self.lst


class TestDumpables(unittest.TestCase):

    def test_state(self):
        """tests DumpableState class"""
        inst = State()
        self.assertTrue(
            is_jsonable(inst.as_dict())
        )

    def test_list(self):
        """tests DumpableList class"""
        inst = List()
        self.assertTrue(
            is_jsonable(inst.as_list())
        )

    def test_casting(self):
        """tests casting function"""
        state_inst = State()
        list_inst = List()
        self.assertTrue(
            is_jsonable(make_iterable_serializable(list_inst))
        )
        self.assertTrue(
            is_jsonable(make_mapping_serializable(val=state_inst))
        )



import time
import unittest
from hein_control.action import ConfiguredAction
from hein_control.scheduler import SamplingScheduler, ExecuteOnTimePoints
from hein_control.mixin import is_jsonable


# logging.basicConfig(level=logging.INFO)


@ConfiguredAction
def sleepy(t):
    time.sleep(t)


@ConfiguredAction
def slow_step():
    sleepy(0.1)


@ConfiguredAction
def fast_step():
    sleepy(0.01)


def as_dict_serializable(action) -> bool:
    """verifies that the as_dict representation is serializable"""
    return is_jsonable(action.as_dict())


class TestScheduler(unittest.TestCase):
    def test_timepoint_addition(self):
        """tests timepoint addition to a scheduler instance"""
        sched = SamplingScheduler('slow_step', fast_step)
        sched.insert_time_point(0.05, number_of_points=4)
        self.assertTrue(as_dict_serializable(sched))
        self.assertEqual(len(sched.time_points), 4)
        # todo
        #   - test relative insertion
        #   - test list insertion
        #   - test list pairs insertion

    def test_timepoint_naming(self):
        """tests naming order of timepoints"""
        sched = SamplingScheduler(fast_step)
        sched.insert_time_point(0.05, number_of_points=4)
        for ind, tp in enumerate(sched.time_points):
            self.assertEqual(
                tp.name,
                f'Time Point #{ind+1}',
                'checking automatic naming'
            )
        sched.insert_time_point(
            0.07,
            relative_to='start',
        )
        for ind, tp in enumerate(sched.time_points):
            self.assertEqual(
                tp.name,
                f'Time Point #{ind+1}',
                'checking automatic name update'
            )
        sched.insert_time_point(
            0.08,
            timepoint_name='manually named'
        )
        counter = 1
        # check that ordering of unnamed points is preserved
        for ind, tp in enumerate(sched.time_points):
            if ind == 2:
                self.assertEqual(
                    tp.name,
                    'manually named',
                    'check preservation of manual name'
                )
            else:
                self.assertEqual(
                    tp.name,
                    f'Time Point #{counter}',
                    'checking automatic name update'
                )
                counter += 1

        sched.insert_time_point(
            0.1,
            number_of_points=3,
            relative_to='previous',
            timepoint_name='test incremental point',
        )
        for ind, tp in enumerate(sched.time_points[-3:]):
            self.assertEqual(
                tp.name,
                f'test incremental point#{ind + 1}',
                'check increment of manual name'
            )
        self.assertTrue(as_dict_serializable(sched))

    def test_execution(self):
        """tests execution of the scheduler"""
        sched = SamplingScheduler(slow_step, 'fast_step')
        sched.insert_time_point(0.5, number_of_points=4)
        self.assertFalse(sched.started)
        sched.start_sequence()
        self.assertTrue(sched.started)
        self.assertFalse(sched.paused)
        sched.pause_sequence()
        self.assertTrue(sched.paused)
        sched.start_sequence()
        while sched.time_points[0].status_code != 3:
            time.sleep(0.1)
        self.assertEqual(sched.completed_time_points, 1, 'one complete time point')
        self.assertGreater(sched.remaining_time_points, 0, 'non-zero points remaining')
        self.assertGreater(sched.triggered_time_points, 0, 'non-zero number triggered')
        self.assertGreater(sched.uncompleted_time_points, 0, 'more than one uncompleted time point')
        self.assertGreater(sched.total_time_points, sched.triggered_time_points, 'total greater than number triggered')
        sched.join()  # wait for scheduler to complete
        self.assertEqual(sched.remaining_time_points, 0, 'no more time points')
        self.assertEqual(sched.triggered_time_points, sched.total_time_points, 'triggered count equals time points')
        self.assertEqual(sched.completed_time_points, sched.total_time_points, 'complete count is accurate')
        self.assertEqual(sched.uncompleted_time_points, 0, 'no uncompleted time points')
        self.assertFalse(sched.busy)
        self.assertTrue(as_dict_serializable(sched))

    def test_triggering(self):
        """tests manual triggering of a scheduler"""
        sched = SamplingScheduler('fast_step')
        sched.trigger()
        self.assertIsNotNone(sched.last_triggered_time_point)
        sched.wait_for_time_point_completion()

    def test_retrieval(self):
        """tests value retrieval from a scheduler"""
        sched = SamplingScheduler('fast_step')
        sched.insert_time_point(0.)
        self.assertIsNone(sched.last_triggered_time_point)
        sched.start_sequence()
        sched.join()
        self.assertIsNotNone(sched.last_triggered_time_point)
        sched.as_dict()

    def test_too_fast(self):
        """tests what happens when time points are executed too fast"""
        sched = SamplingScheduler('slow_step', fast_step)
        sched.insert_time_point(0.001, number_of_points=4)
        sched.start_sequence()
        sched.trigger()
        sched.join()

    def test_duplicate_naming(self):
        """tests catch and acceptance of duplicate configuration names"""
        prev_name = ''
        for i in range(10):
            sched = SamplingScheduler(
                sequence_name='convenience name'
            )
            self.assertNotEqual(prev_name, sched.sequence.name)
            prev_name = sched.sequence_name

    def test_trigger_start(self):
        """tests an edge case error where a scheduler is triggered and then started"""
        sched = SamplingScheduler(slow_step, slow_step)
        sched.trigger()
        sched.start_sequence()

    # def test_slow_spoolup(self):
    #     """tests handling of ActionTimePoint spool up situations"""
    #     # todo figure out how to test slow spool up situations (execution iteration is faster than _triggered set)
    #     sched = SamplingScheduler(fast_step)
    #     sched.insert_time_point(0.)
    #     tp = sched.time_points[0]
    #     tp._executor = slow_spool_executor
    #     sched.start_sequence()
    #     tp._triggered = True
    #     sched.join()

    # def test_addition_actions(self):  # todo
    #     pass

    # def test_action_override(self):  # todo
    #     pass


class TestTimePointConditional(unittest.TestCase):
    @staticmethod
    def dummy():
        return 'value'

    def test_conditional(self):
        """tests the conditional which runs actions on specified timepoints"""

        tp_nums = [1, 3, 5]
        condl_action = ExecuteOnTimePoints(
            self.dummy,
            *tp_nums,
        )
        sched = SamplingScheduler(
            fast_step,
            condl_action,
        )

        sched.insert_time_point(0.2, number_of_points=10, relative_to='previous')
        sched.start_sequence()
        sched.join()
        for ind, timepoint in enumerate(sched.time_points):

            if ind + 1 in tp_nums:
                expected = 'value'
            else:
                expected = None
            self.assertEqual(
                timepoint.action_returns['dummy'],
                expected,
                msg=f'ensure timepoint {ind + 1} conditional execution was correct'
            )



"""
DEPRECATED: use the `.sequencing.automation` module instead.

Module for managing automation sequences (series of Steps).
"""
import warnings
from .sequencing.automation import Automation, build_automation_sequence

warnings.warn(  # v6
    'the automation module has been moved into the sequencing subpackage, please update your imports',
    DeprecationWarning,
    stacklevel=2,
)



"""module for managing component states"""

import time
import logging

from typing import Union, MutableMapping, List, Callable, Type
from enum import IntEnum
from functools import wraps

from .mixin.reg import InstanceRegistry


logger = logging.getLogger(__name__)


class OperationalState(IntEnum):
    """
    operational state tracker for components. By convention, use 0 to indicate offline, negative values for
    error states, and positive values for online states
    """
    ERROR = -1
    OFFLINE = 0
    ONLINE = 1


class GenericActionState(IntEnum):
    """
    Basic action state tracker for component. By convention, use 0 to represent an inactive state for the inactive
    state check decorator to function.
    """
    IDLE = 0
    ACTIVE = 1


class BadComponentState(Exception):
    def __init__(self, *components: 'ComponentState', threshold_state: IntEnum):
        """
        Exception signifying that bad state was encountered on the provided components

        :param components: components which have a bad state
        :param threshold_state: threshold state
        """
        super(BadComponentState, self).__init__(
            f'a bad state was encountered with the component{"s" if len(components) > 0 else ""}: ' +
            ', '.join([
                f'{component.component_name}: {component.current_operational_state.name}'
                for component in components
            ])
            + f'; threshold state: {threshold_state.name}'
        )


def ensure_component_online(fn: Callable):
    """class instance decorator which will check for an online state before executing the decorated function"""
    @wraps(fn)
    def decorated(self: 'ComponentState', *args, **kwargs):
        if self.current_operational_state <= 0:
            logger.error(
                f'the component {self.component_name} is in operational state {self.current_operational_state.name} '
                f'(required: {OperationalState.ONLINE}) '
                f'when calling {fn.__name__}'
            )
            raise BadComponentState(self, threshold_state=OperationalState.OFFLINE)
        return fn(self, *args, **kwargs)

    return decorated


def ensure_component_inactive(fn: Callable):
    """
    class instance decorator which will check that the component is inactive before
    executing the decorated function
    """
    @wraps(fn)
    def decorated(self: 'ComponentState', *args, **kwargs):
        if self.current_action_state != 0:
            # retrieve the expected idle state enumeration
            expected_idle_state = self.current_action_state.__class__(0)
            logger.error(
                f'the component {self.component_name} is in action state {self.current_action_state.name} '
                f'(required: {expected_idle_state.name}) when calling {fn.__name__}'
            )
            raise BadComponentState(self, threshold_state=expected_idle_state)
        return fn(self, *args, **kwargs)

    return decorated


class ComponentState(InstanceRegistry):

    # dedicated component-specific mapping
    _components: MutableMapping[str, 'ComponentState'] = {}

    # override this when subclassing to indicate more advanced component action states
    action_states: Type[IntEnum] = GenericActionState

    # override this when subclassing to indicate more advanced component operational states
    #   negative values indicate error states, 0 is offline, and positive is an operational state
    operational_states: Type[IntEnum] = OperationalState

    def __init__(self,
                 component_name: str,
                 current_operational_state: Union[int, str, IntEnum] = None,
                 current_action_state: Union[int, str, IntEnum] = None
                 ):
        """
        ABC for tracking a component's operational state

        :param component_name: reference name for the component being registered
        :param current_operational_state: current state of the component
        """
        # todo create an InstanceRegistry subclass which tracks subclass instances (i.e. all components)
        if component_name in ComponentState.class_registry:
            raise NameError(
                f'the component "{component_name}" is already defined, please retrieve using get_component_by_name or '
                f'choose a different name for the component'
            )
        InstanceRegistry.__init__(self)
        ComponentState._components[component_name] = self
        self.component_name = component_name
        self._current_operational_state: OperationalState = None
        self.current_operational_state = current_operational_state or 0
        self._current_action_state: IntEnum = None
        self.current_action_state = current_action_state or 0
        self.logger = logger.getChild(self.component_name)

    @staticmethod
    def _match_instance_key(inst: "ComponentState", identifier: str) -> bool:
        """matches the provided identifier against the component name and standard identifiers"""
        if inst._uuid == identifier:
            return True
        elif inst.uuid == identifier:
            return True
        elif identifier == inst.component_name:
            # todo consider matching case-insensitive
            return True
        return False

    @property
    def current_operational_state(self) -> IntEnum:
        """The current state of the component"""
        return self._current_operational_state

    @current_operational_state.setter
    def current_operational_state(self, value: Union[int, str, IntEnum]):
        self.update_operational_state(value)

    def get_operational_state_from_key(self, value: Union[int, str, IntEnum]) -> IntEnum:
        """
        Gets an operational state from the provided key. If an IntEnum instance which is not the same
        class as the assigned operational state class of the instance, the equivalent state will be retrieved.

        :param value: value, name, or IntEnum
        """
        if isinstance(value, self.operational_states):
            return value
        elif isinstance(value, IntEnum):
            try:
                return self.operational_states(value.value)
            except ValueError as e:
                raise ValueError(f'the operational state {value} does not have an equivalent {self.operational_states.__name__} value')
        elif type(value) is str:
            return getattr(self.operational_states, value.upper())
        elif type(value) is int:
            return self.operational_states(value)
        else:
            raise TypeError(f'the type {type(value)} is not supported for a current component operational state')

    def update_operational_state(self, value: Union[int, str, IntEnum]) -> IntEnum:
        """
        Updates the current state of the Component

        :param value: value to set the state to
        :return: component state
        """
        self._current_operational_state = self.get_operational_state_from_key(value)
        return self._current_operational_state

    @property
    def current_action_state(self) -> IntEnum:
        """the current action state of the component"""
        return self._current_action_state

    @current_action_state.setter
    def current_action_state(self, value: Union[int, str, IntEnum]):
        self.update_action_state(value)

    def update_action_state(self, value: Union[int, str, IntEnum]) -> IntEnum:
        """
        Updates the current action state of the component

        :param value: value to set the state to
        :return: component action state
        """
        if type(value) is int:
            value = self.action_states(value)
        elif type(value) is str:
            value = getattr(self.action_states, value.upper())
        elif isinstance(value, self.action_states) is False:
            raise TypeError(f'the type {type(value)} is not supported for a current component action state')
        self._current_action_state = value
        return self._current_action_state

    @classmethod
    def check_component_states(cls,
                               threshold_state: Union[int, OperationalState],
                               *components: Union[str, 'ComponentState'],
                               ) -> List['ComponentState']:
        """
        Checks the provided components against the minimum component state value, returning any components which do not
        pass the check.

        :param threshold_state: Minimum state value (components with states lower than this will fail the check)
        :param components: components to check
        :return: list of failing components
        """
        components = cls.get_components_to_monitor(*components)
        return [
            component
            for component in components
            if component.current_operational_state <= threshold_state
        ]

    @classmethod
    def check_component_states_raise(cls,
                                     threshold_state: Union[int, OperationalState],
                                     *components: 'ComponentState',
                                     ):
        """
        Checks the list of components against the minimum state and raises an error if any fail the test

        :param threshold_state: the threshold state to monitor for
        :param components: components to check
        """
        bad_states = cls.check_component_states(threshold_state, *components)
        if len(bad_states) > 0:
            logger.error(
                f'a bad state was encountered on the following component{"s" if len(bad_states) == 1 else ""}: '
                + ", ".join([
                    f'{component.component_name} {component.current_operational_state.value} '
                    f'({component.current_operational_state.name})'
                    for component in bad_states
                ])
            )
            raise BadComponentState(*bad_states, threshold_state=threshold_state)

    @classmethod
    def components_from_names(cls, *components: Union[str, 'ComponentState']) -> List['ComponentState']:
        """
        Retrieves a list of ComponentState subclasses from the provided list.

        :param components: components to find
        :return: list of ComponentState instances
        """
        out = []
        for component in components:
            if isinstance(component, ComponentState) is True:
                out.append(component)
            elif type(component) is str:
                out.append(cls._components[component])
            else:
                raise TypeError(f'the component "{component}" is not recognized')
        return out

    @classmethod
    def get_components_to_monitor(cls, *components: Union[str, 'ComponentState']) -> List['ComponentState']:
        """
        Gets a list of components to monitor from the provided arguments. If no components are provided, the complete
        list of components is provided.

        :param components: components to monitor
        :return: list of ComponentState instances
        """
        if len(components) > 0:
            return cls.components_from_names(*components)
        return [component for _, component in cls._components.items()]

    @classmethod
    def components_state_monitor_sleep(cls,
                                       duration: float,
                                       *components: Union[str, 'ComponentState'],
                                       cycle_time: float = 0.1,
                                       threshold_state: Union[int, IntEnum] = OperationalState.OFFLINE,
                                       ):
        """
        Executes a sleep while also monitoring the state of the specified components. If no components are provided,
        all components will be monitored.

        :param duration: duration to sleep (approximate)
        :param components: components to monitor for a bad state. Can be either the name of the component or the
            ComponentState instance itself.
        :param cycle_time: cycle frequency (s; a check for a bad state will be executed every this seconds)
        :param threshold_state: the state to monitor for (states less or equal to this will raise an error)
        """
        # retrieve components as specified
        monitored_components = cls.get_components_to_monitor(*components)
        logger.info(
            f'waiting for ~{duration} s while monitoring '
            + f'{len(monitored_components)} components'
            if len(monitored_components) > 0
            else monitored_components[0].component_name
        )
        end_time = time.time() + duration
        while time.time() < end_time:
            cls.check_component_states_raise(threshold_state, *monitored_components)
            time.sleep(cycle_time)
        logging.info(f'finished waiting, continuing')

    def state_monitor_sleep(self,
                            duration: float,
                            cycle_time: float = 0.1,
                            threshold_state: Union[int, IntEnum] = OperationalState.OFFLINE,
                            ):
        """
        Executes a sleep while also monitoring the state of the component. To monitor the state of multiple components,
        use components_state_monitor_sleep.

        :param duration: duration to sleep (approximate)
        :param cycle_time: cycle frequency (s; a check for a bad state will be executed every this seconds)
        :param threshold_state: the state to monitor for (states less or equal to this will raise an error)
        """
        return self.components_state_monitor_sleep(
            duration,
            self.component_name,
            cycle_time=cycle_time,
            threshold_state=threshold_state,
        )

    @classmethod
    def components_state_monitor_flag_check(cls,
                                            flag_check: Callable,
                                            *components: Union[str, 'ComponentState'],
                                            target_flag_state=True,
                                            cycle_time: float = 0.1,
                                            threshold_state: Union[int, IntEnum] = OperationalState.OFFLINE,
                                            **flag_check_kwargs
                                            ):
        """
        Monitors the provided flag check while also monitoring the state of the specified components. If no components
        are provided, all components will be monitored.

        :param flag_check: method which will return a value that will be checked against the target_flag_state
        :param components: components to monitor for a bad state. Can be either the name of the component or the
            ComponentState instance itself.
        :param cycle_time: cycle frequency (s; a check for a bad state will be executed every this seconds)
        :param threshold_state: the state to monitor for (states less or equal to this will raise an error)
        :param target_flag_state: target flag state to monitor for
        :param flag_check_kwargs: keyword arguments for the flag check
        """
        # retrieve components as specified
        monitored_components = cls.get_components_to_monitor(*components)
        logger.info(
            f'waiting for {flag_check.__name__} to be {target_flag_state} while monitoring '
            + f'{len(monitored_components)} components' if len(monitored_components) > 0
            else monitored_components[0].component_name
        )
        while flag_check(**flag_check_kwargs) != target_flag_state:
            cls.check_component_states_raise(threshold_state, *monitored_components)
            time.sleep(cycle_time)
            # todo add timeout
        logging.info(f'finished waiting, continuing')

    def state_monitor_flag_check(self,
                                 flag_check: Callable,
                                 target_flag_state=True,
                                 cycle_time: float = 0.1,
                                 threshold_state: Union[int, IntEnum] = OperationalState.OFFLINE,
                                 **flag_check_kwargs
                                 ):
        """
        Monitors the state of the component, waiting for the threshold state. To monitor multiple components, use
        components_state_monitor_flag_check.


        :param flag_check: method which will return a value that will be checked against the target_flag_state
        :param cycle_time: cycle frequency (s; a check for a bad state will be executed every this seconds)
        :param threshold_state: the state to monitor for (states less or equal to this will raise an error)
        :param target_flag_state: target flag state to monitor for
        :param flag_check_kwargs: keyword arguments for the flag check
        """
        self.components_state_monitor_flag_check(
            flag_check,
            self.component_name,
            target_flag_state=target_flag_state,
            cycle_time=cycle_time,
            threshold_state=threshold_state,
            **flag_check_kwargs,
        )

    @classmethod
    def get_component_by_name(cls, component_name: str) -> 'ComponentState':
        """
        Retrieves a component by its name. If you need to retrieve a component that you know is already instantiated,
        use this method.

        :param component_name: name assigned to the component
        """
        try:
            return cls._components[component_name]
        except KeyError:
            raise KeyError(f'the component "{component_name}" is not a defined component')

    @classmethod
    def get_or_create_component(cls, *args, component_name: str = None, **kwargs) -> 'ComponentState':
        """
        Retrieves an existing component or instantiates it as needed. This is intended to be the primary factory for
        component instantiation.

        :param args: component arguments
        :param component_name: name for the component
        :param kwargs: component keyword arguments
        :return: instantiated components
        """
        if component_name is None:
            raise ValueError(f'a component name must be specified')
        if component_name in cls._components:
            return cls._components[component_name]
        return cls(
            *args,
            component_name=component_name,
            **kwargs,
        )

    def wait_for_component_operational_state(self,
                                             desired_state: Union[int, IntEnum, str],
                                             cycle_time: float = 0.1,
                                             timeout: float = None,  # todo implement timeout for wait
                                             ):
        """
        Waits for the specified components to be at or above the specified operational state.

        :param desired_state: desired state for the component to be in
        :param cycle_time: state check cycle time
        """
        # todo create a generic method to wait for multiple components
        desired_state = self.get_operational_state_from_key(desired_state)
        if timeout is not None:
            max_time = time.time() + timeout
        self.logger.info(f'waiting for {desired_state.name} operational state')
        while self.current_operational_state < desired_state:
            if timeout is not None and time.time() > max_time:
                self.logger.error(
                    f'timeout of {timeout} s elapsed while waiting for {self.component_name} to be {desired_state.name}'
                )
            time.sleep(cycle_time)



import logging
import time
import unittest
from hein_control.action import Action, EvalAction, TrackedDecision, ConfiguredAction, ConfiguredDecision, \
    ThreadedTrackedAction, TrackedAction, ConfiguredActionList, ConfigurableInstance
from hein_control.action.operators import execute_sequence_actions, run_automation, decision
from hein_control.mixin import Configuration, Status, is_jsonable

logger = logging.getLogger(__name__)

one_second_return = 'abcde'
fifty_return = [1, 2, 3, 4, 5]


def one_second():
    time.sleep(1)
    return one_second_return


def fifty_ms():
    time.sleep(0.05)
    return fifty_return


def user_time(dur, **kwargs):
    time.sleep(dur)
    return kwargs


def five_second():
    time.sleep(5.)
    return fifty_return

def return_true(*args, **kwargs) -> bool:
    return True

def return_false(*args, **kwargs) -> bool:
    return False

error_details = 'you did something horrific and it broke the code'


def erronious():
    raise ValueError(error_details)


def passthrough(*args, **kwargs):
    return args, kwargs


def as_dict_serializable(action) -> bool:
    """verifies that the as_dict representation is serializable"""
    return is_jsonable(action.as_dict())


class TestActions(unittest.TestCase):
    def setUp(self) -> None:
        # register all actions
        Action.register_action(one_second)
        Action.register_action(fifty_ms)
        Action.register_action(user_time)

    def test_registry(self):
        """tests registry functionality of SchedulerAction"""
        self.assertTrue(
            Action.action_is_registered(one_second)
        )
        self.assertTrue(
            Action.action_is_registered(fifty_ms)
        )

    def test_basic_action(self):
        """test functionality of action calling and return"""
        action = Action.register_action(one_second)
        self.assertEqual(  # test name attribute
            action.name,
            one_second.__name__
        )
        ret = action()  # execute action, test return
        self.assertEqual(
            one_second_return,
            ret
        )

    def test_argument_passthrough(self):
        """tests argument and kwarg passthrough"""
        action = Action.register_action(passthrough)
        args = (0.123, 123, 456)
        kwargs = {'asdf': 'qwerty'}
        ret = action(
            *args,
            **kwargs
        )
        self.assertEqual(
            (args, kwargs),
            ret
        )

    def test_props(self):
        """tests properties of trackedaction"""
        pass  # todo

    def test_duplication(self):
        """tests duplication of TrackedAction"""
        act = TrackedAction(passthrough)
        act2 = act.copy_configuration()
        act()
        self.assertRaises(  # verify that action cannot be executed twice
            RuntimeError,
            act,
        )
        act2()  # ... and that "copy" instance still works

    def test_threaded(self):
        """test threaded tracked action"""
        act = ThreadedTrackedAction(user_time)
        self.assertFalse(act.triggered)
        self.assertFalse(act.started)
        act(dur=0.2)
        self.assertTrue(act.triggered)
        self.assertTrue(act.in_progress)
        act.wait_for_completion()
        self.assertTrue(act.complete)

    def test_arg_kwarg(self):
        """tests argument and kwarg specification and passthrough"""
        args = (0.123, 123, 456)
        kwargs = {'asdf': 'qwerty'}
        # test argument updating
        action = TrackedAction(
            passthrough,
            *args,
            **kwargs
        )
        self.assertEqual(
            action.args,
            args
        )
        self.assertEqual(
            action.kwargs,
            kwargs
        )
        action(
            789,
            asdf='al;skdjf;'
        )
        self.assertEqual(
            action.args,
            (789, 123, 456)
        )
        self.assertEqual(
            action.kwargs,
            {'asdf': 'al;skdjf;'}
        )

    def test_next_action_kwarg(self):
        """tests passing a next action"""
        args = (0.123, 123, 456)
        kwargs = {'asdf': 'qwerty'}
        config_1 = ConfiguredAction(
            passthrough,
            *args,
            **kwargs
        )
        tracked = TrackedAction(ConfiguredAction)
        args2 = (1, 2)
        kwargs2 = {'squirrels': 'cool'}
        config_2 = ConfiguredAction(
            passthrough,
            *args2,
            **kwargs2
        )
        self.assertIsNone(
            config_1.next_action
        )
        self.assertIsNone(
            tracked.next_action
        )
        config_1.next_action = config_2
        self.assertEqual(
            config_1.next_action,
            config_2,
        )

    def test_error_catching(self):
        """tests that errors are properly raised"""
        action = Action.register_action(erronious)
        self.assertRaises(
            ValueError,
            action
        )
        self.assertRaises(
            TypeError,
            Action,
            ('asdf123'),
            'check that non-callables are correctly rejected as actions'
        )


def return_str() -> str:
    return 'metal'


def return_true() -> bool:
    return True


def return_false() -> bool:
    return False


class ConditionClass:
    flag = False


class TestConfiguredAction(unittest.TestCase):
    def test_instance_by_id(self):
        # todo docstrings suggest that only the UUID can be used
        conf_action_1 = ConfiguredAction(passthrough,
                                         configuration_name='config_test_by_id',)
        self.assertRaises(
            NameError,
            ConfiguredAction.class_instance_by_id,
            ('this_should_be_undefined'),
        )

    def test_serializable(self):
        """tests that the registered actions are serializable"""
        conf_action_apple = ConfiguredAction(passthrough,
                                             configuration_name='apple')
        self.assertTrue(as_dict_serializable(conf_action_apple))

    def test_configured_action(self):
        """test all the methods of the configured action class"""
        args = ('metal', 'squirrel')
        kwargs = {'squirrels': 'are cute'}
        conf_action_apple = ConfiguredAction(passthrough,
                                             configuration_name='apple',
                                             *args,
                                             **kwargs)
        conf_action_orange = conf_action_apple.duplicate_configuration(new_name='orange')
        self.assertEqual(
            f'0: {Status.ACTION_STATES[0]}',
            conf_action_apple.status,
        )
        self.assertEqual(conf_action_apple.name,
                         'apple')
        self.assertEqual(conf_action_orange.name,
                         'orange')

        self.assertEqual(conf_action_apple.next_action
                         , None)
        self.assertEqual(conf_action_orange.previous_actions,
                         [])

        conf_action_apple.next_action = conf_action_orange
        self.assertEqual(conf_action_apple.next_action,
                         conf_action_orange)
        self.assertEqual(conf_action_orange.previous_actions,
                         [conf_action_apple])

        self.assertEqual(ConfiguredAction.class_instance_by_id(conf_action_apple.uuid),
                         conf_action_apple)
        self.assertEqual(ConfiguredAction.search_class_for_id(conf_action_apple.uuid, 'ConfiguredAction'),
                         conf_action_apple)
        # todo throws a  TrackedAction error for some reason when running all the unit tests, but passes when only
        #  this test is run
        self.assertIs(ConfiguredAction.get_instance_by_uuid(conf_action_apple.uuid),
                      conf_action_apple)

        self.assertEqual(ConfiguredAction.ensure_configuration(conf_action_apple),
                         conf_action_apple)
        self.assertEqual(ConfiguredAction.ensure_configuration(conf_action_apple),
                         conf_action_apple)
        self.assertEqual(ConfiguredAction.ensure_configuration(conf_action_apple.name),
                         conf_action_apple)
        conf_action_pear = ConfiguredAction.ensure_configuration(passthrough)
        self.assertEqual(conf_action_pear,
                         ConfiguredAction.class_instance_by_id(conf_action_pear.uuid))
        self.assertIs(ConfiguredAction.ensure_configuration(one_second),
                      ConfiguredAction.class_instance_by_id('one_second'))
        self.assertRaises(
            ValueError,
            ConfiguredAction.ensure_configuration,
            ('not a registered action'),
        )

        tracked = conf_action_apple.get_tracked_from_config()
        self.assertEqual(conf_action_apple,
                         tracked.configuration)
        self.assertEqual(TrackedAction,
                         type(tracked))

    def test_conditional_configured_action(self):
        """tests the conditional configured action setup"""
        flag = False
        def evaluator():
            return flag

        cond = ConfiguredAction(return_str, condition=evaluator)
        track_1 = cond.get_tracked_from_config()
        track_1()
        self.assertEqual(track_1.status_code, 3)
        self.assertIsNone(track_1.action_return, "ensure that the action didn't run")
        flag = True
        track_2 = cond.get_tracked_from_config()
        track_2()
        self.assertEqual(track_2.status_code, 3)
        self.assertEqual(track_2.action_return, 'metal', "ensure that the action executed")
        output = track_2.as_dict()
        self.assertEqual(output['condition_evaluation'], True, "ensure output to dictionary")

    def test_conditional_configured_accession(self):
        """tests the configuredaction condition where the condition is a class flag"""
        cond_class = ConditionClass()
        cond_conf_1 = ConfiguredAction(return_str, condition=(cond_class, 'flag'))
        track_1 = cond_conf_1.get_tracked_from_config()
        track_1()
        self.assertEqual(track_1.status_code, 3)
        self.assertIsNone(track_1.action_return, "ensure that the action didn't run")
        cond_class.flag = True
        track_2 = cond_conf_1.get_tracked_from_config()
        track_2()
        self.assertEqual(track_2.status_code, 3)
        self.assertEqual(track_2.action_return, 'metal', "ensure that the action executed")


class TestEvalAction(unittest.TestCase):
    def setUp(self) -> None:
        Action.register_action(return_str)
        Action.register_action(return_false)
        EvalAction.register_action(return_true)

    def test_evalaction(self):
        """tests the evalaction class"""
        str_action = ConfiguredAction(Action.register_action('return_str'))
        self.assertFalse(str_action.action.returns_bool)

        true_action = ConfiguredAction(return_true, next_action=str_action)
        self.assertTrue(true_action.action.returns_bool)

        false_action_1 = ConfiguredAction(Action.register_action('return_false'))
        self.assertTrue(false_action_1.action.returns_bool)

        false_action_2 = ConfiguredAction(false_action_1)
        self.assertTrue(false_action_2.action.returns_bool)


class TestDecisions(unittest.TestCase):
    @staticmethod
    def print_false():
        print('this should not be printed')

    @staticmethod
    def print_true():
        print('true')

    @staticmethod
    def print_also_true():
        print('also true')

    @staticmethod
    def return_true(string, **kwargs) -> bool:
        return True

    @staticmethod
    def return_false(string, **kwargs) -> bool:
        return False

    def test_decisionaction(self):
        """test ConfiguredDecision and TrackedDecision classes"""
        # create short sequence/decision tree of actions
        false_next = ConfiguredAction(self.print_false)
        true_next = ConfiguredAction(self.print_true)
        true_eval_action = EvalAction(self.return_true)
        false_eval_action = EvalAction(self.return_false)
        args = ('function returns',)
        true_kwargs = {'next_action': true_next}
        true_action = ConfiguredAction(
            true_eval_action,
            *args,
            **true_kwargs,
        )
        false_kwargs = {'next_action': false_next}
        false_action = ConfiguredAction(
            false_eval_action,
            *args,
            **false_kwargs,
        )
        config_decision = ConfiguredDecision(
            false_action,
            true_action,
            configuration_name='logic checklist'
        )
        # create the decision action and execute it
        decision_action = TrackedDecision(configuration=config_decision)
        self.assertTrue(as_dict_serializable(decision_action))
        decision_action()
        self.assertTrue(as_dict_serializable(decision_action))
        decision_next_action = decision_action.next_action
        self.assertEqual(decision_next_action, true_next)
        # change the next action of true action, make a new decision action and make sure the new next action gets
        # returned correctly
        also_true = ConfiguredAction(self.print_also_true)
        true_action.next_action = also_true
        # ensure that attempts to create a TrackedAction subclass with TrackedAction as the action will error
        self.assertRaises(
            NameError,
            TrackedDecision,
            decision_action
        )
        decision_2 = decision_action.copy_configuration()
        decision_2()
        self.assertEqual(decision_2.action_return, also_true)


class TestTrackedAction(unittest.TestCase):
    def test(self):
        """tests tracked action class"""
        tracked_action = TrackedAction(
            one_second
        )
        self.assertTrue(as_dict_serializable(tracked_action))
        start_time = time.time()
        tracked_action()
        self.assertTrue(as_dict_serializable(tracked_action))
        end_time = time.time()
        self.assertAlmostEqual(
            start_time,
            tracked_action.time_started,
            places=1,
        )
        self.assertAlmostEqual(
            end_time,
            tracked_action.time_completed,
            places=1,
        )
        self.assertAlmostEqual(
            end_time - start_time,
            tracked_action.action_duration,
            places=1,
            )
        self.assertEqual(
            tracked_action.configuration.action.name,
            one_second.__name__
        )
        self.assertEqual(
            tracked_action.action_return,
            one_second_return
        )
        dct = tracked_action.as_dict()  # todo more elaborate test for dict testing

        configured_action_2 = ConfiguredAction(passthrough,
                                               next_action=tracked_action.configuration)
        tracked_action_2 = TrackedAction(
            configured_action_2,
        )
        self.assertEqual(
            f'1: {Status.ACTION_STATES[1]}',
            tracked_action_2.status,
        )
        self.assertEqual(configured_action_2,
                         tracked_action_2.configuration)
        self.assertEqual(configured_action_2.action,
                         tracked_action_2.configuration.action)
        self.assertEqual(configured_action_2.next_action,
                         tracked_action_2.next_action)
        self.assertEqual(ConfiguredAction.ensure_configuration('one_second'),
                         tracked_action_2.next_action)

        tracked_action_3 = tracked_action_2.copy_configuration()
        self.assertEqual(tracked_action_3.configuration, tracked_action_2.configuration)
        self.assertEqual(tracked_action_3.args, tracked_action_2.args)
        self.assertEqual(tracked_action_3.kwargs, tracked_action_2.kwargs)

        self.assertEqual((),
                         tracked_action_2.args)
        args = ('squirrels', )
        tracked_action_2.update_args(*args)
        self.assertEqual(args,
                         tracked_action_2.args)
        kwargs = {'squirrels': 'almonds'}
        tracked_action_2.update_kwargs(**kwargs)
        self.assertEqual(kwargs,
                         tracked_action_2.kwargs)
        tracked_action_2()
        self.assertEqual(
            f'3: {Status.ACTION_STATES[3]}',
            tracked_action_2.status,
        )
        self.assertEqual(True, tracked_action_2.triggered)

        self.assertNotEqual(tracked_action_2.args, tracked_action.args)
        self.assertNotEqual(tracked_action_2.kwargs, tracked_action.kwargs)

    def test_error_catching(self):
        action = Action.register_action(erronious)
        action = TrackedAction(action)
        action()
        self.assertTrue(action.error)
        self.assertEqual(action.error_details.args[0], error_details)
        self.assertEqual(-1, action.status_code)


class TestConfiguration(unittest.TestCase):
    def test_configuration(self):
        """tests the configuration class"""
        test = Configuration(1, 2, 3, a='b', asdf='c')
        self.assertEqual(
            test.args,
            (1, 2, 3)
        )
        test.update_args(2, 3, 4)
        self.assertEqual(
            test.args,
            (2, 3, 4)
        )
        self.assertEqual(
            test.kwargs,
            {'a': 'b', 'asdf': 'c'}
        )
        test.update_kwargs(a='c', ghjk=2)
        self.assertEqual(
            test.kwargs,
            {'a': 'c', 'asdf': 'c', 'ghjk': 2}
        )
        test.freeze = True
        self.assertRaises(
            ValueError,
            test.update_args,
            1
        )


class TestLogic(unittest.TestCase):
    def test_logic(self):
        true_config = ConfiguredAction(return_true)
        false_config = ConfiguredAction(return_false)
        configured_decision = ConfiguredDecision(false_config, true_config)
        # self.assertEqual(  # todo add comparison magic methods to ActionList and derivatives
        #     configured_decision.logic_sequence,
        #     ConfiguredActionList(false_config, true_config)
        # )

        tracked_1 = TrackedDecision(configured_decision)
        tracked_2 = configured_decision.get_tracked_from_config()
        self.assertEqual(
            tracked_1.configuration,
            tracked_2.configuration
        )
        self.assertEqual(
            tracked_1.configuration(),
            tracked_2.configuration()
        )

        kwargs = {'squirrel': 'almonds'}
        tracked_2.update_kwargs(**kwargs)
        self.assertEqual(
            tracked_2.kwargs,
            kwargs
        )
        tracked_3 = configured_decision.get_tracked_from_config(**kwargs)
        self.assertEqual(
            tracked_3.kwargs,
            kwargs
        )

        next_conf = ConfiguredAction(passthrough)
        true_config.next_action = next_conf
        tracked_4 = true_config.get_tracked_from_config()
        self.assertEqual(
            tracked_4.next_action,
            next_conf
        )


class TestOperators(unittest.TestCase):
    def log_1(self, one: str, **kwargs):
        logger.debug(one)
        logger.debug(kwargs)
        return one, kwargs

    def log_2(self, two: str, **kwargs):
        logger.debug(two)
        logger.debug(kwargs)
        return two, kwargs

    def return_bool(self, bool: bool) -> bool:
        return bool

    def test_execute_sequence_actions(self):
        config_log_1 = ConfiguredAction(self.log_1)
        config_log_2 = ConfiguredAction(self.log_2)
        tracked_1 = config_log_1.get_tracked_from_config()
        tracked_2 = config_log_2.get_tracked_from_config()
        kwargs = {'log_1': {'one': 'one!',
                            'kwargs': 'kwargs for one'},
                  'log_2': {'two': 'two!',
                            'kwargs': 'kwargs for two'}
                  }

        tracked_actions = [tracked_1, tracked_2]
        execute_sequence_actions(*tracked_actions, **kwargs)
        self.assertEqual(False,
                         tracked_1.error,
                         )
        self.assertEqual(False,
                         tracked_2.error,
                         )
        self.assertEqual(
            ('one!', {'kwargs': 'kwargs for one'}),
            tracked_1.action_return)
        self.assertEqual(
            ('two!', {'kwargs': 'kwargs for two'}),
            tracked_2.action_return)

        # cannot execute a tracked action more than once
        self.assertRaises(
            RuntimeError,
            execute_sequence_actions,
            *tracked_actions, **kwargs,
        )

    def test_decision(self):
        config_log_1 = ConfiguredAction(self.log_1)
        config_log_2 = ConfiguredAction(self.log_2)

        eval_action = EvalAction(self.return_bool)
        conf_true = ConfiguredAction(eval_action, next_action=config_log_1, **{'bool': True})
        conf_false = ConfiguredAction(eval_action, next_action=config_log_2, **{'bool': False})

        tracked_true = conf_true.get_tracked_from_config()
        tracked_false = conf_false.get_tracked_from_config()
        logic_checks = [tracked_false, tracked_true]

        # todo should there be a way to pass kwargs to a decision or not?
        decision_next = decision(*logic_checks)
        self.assertTrue(tracked_true.triggered)
        self.assertTrue(tracked_false.triggered)
        self.assertEqual(config_log_1,
                         decision_next)

    def test_run_automation(self):
        config_log_1 = ConfiguredAction(self.log_1, configuration_name='uta')
        config_log_2 = ConfiguredAction(self.log_2, configuration_name='eluveitie')

        first_action = ConfiguredAction(self.log_1,
                                        configuration_name='first action',
                                        **{'one': 'first one'}
                                        )

        eval_action = EvalAction.register_action(self.return_bool)
        conf_true = ConfiguredAction(eval_action, next_action=config_log_1, **{'bool': True})
        conf_false = ConfiguredAction(eval_action, next_action=config_log_2, **{'bool': False})

        configured_decision = ConfiguredDecision(conf_false, conf_true)
        first_action.next_action = configured_decision

        kwargs = {'uta': {'one': 'one!',
                          'kwargs': 'kwargs for one'},
                  'eluveitie': {'two': 'two!',
                                'kwargs': 'kwargs for two'}
                  }
        executed_actions = run_automation(first_action, **kwargs)
        self.assertEqual(3,
                         len(executed_actions))

        self.assertTrue(first_action.status)
        tracked_first_action = executed_actions[0]
        tracked_decision = executed_actions[1]
        tracked_conf_log_1 = executed_actions[2]  # this was the next action from the decision
        self.assertEqual(('first one', {}),
                         tracked_first_action.action_return)
        self.assertEqual(config_log_1,
                         tracked_decision.action_return)
        self.assertEqual(('one!', {'kwargs': 'kwargs for one'}),
                         tracked_conf_log_1.action_return)


class RegMethodsInstance(ConfigurableInstance):
    methods_to_register = [
        'inst_action_one',
        'inst_action_two',
    ]

    def __init__(self):
        ConfigurableInstance.__init__(self, instance_specific_actions=True)

    def inst_action_one(self):
        return 'banana'

    def inst_action_two(self):
        return 'split'


class TestConfigurableInstance(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.test_inst = RegMethodsInstance()
        cls.test_inst2 = RegMethodsInstance()

    def test_registered(self):
        """tests that methods are registered"""
        self.assertTrue(isinstance(self.test_inst.inst_action_one, ConfiguredAction))
        self.assertTrue(isinstance(self.test_inst.inst_action_two, ConfiguredAction))

    def test_blocking_exec(self):
        """tests blocking execution"""
        def execute_and_test(action, result):
            ta = self.test_inst.execute_action(action)
            self.assertEqual(ta.status_code, 3)
            self.assertEqual(ta.action_return, result)
        execute_and_test('inst_action_one', 'banana')
        execute_and_test('inst_action_two', 'split')

    def test_nonblocking_exec(self):
        """tests threaded execution of actions"""
        def execute_and_test(action, result):
            tta = self.test_inst.execute_action_threaded(action)
            tta.wait_for_completion()
            self.assertEqual(tta.status_code, 3)
            self.assertEqual(tta.action_return, result)

        execute_and_test('inst_action_one', 'banana')
        execute_and_test('inst_action_two', 'split')

    def test_method_associations(self):
        """tests that registered methods of a class are not methods of the other"""
        self.assertFalse(
            self.test_inst.inst_action_one is self.test_inst2.inst_action_one,
            "ensure configured actions are not duplicates"
        )
        self.assertFalse(
            self.test_inst.inst_action_one.action.action
            is self.test_inst2.inst_action_one.action.action,
            "ensure actions point to correct instance"
        )



from .serialization import DumpableState


class Status(DumpableState):
    # dictionary of action states
    ACTION_STATES = {
        -1: 'ERROR',
        0: 'CONFIGURED',
        1: 'PENDING',
        2: 'EXECUTING',
        3: 'COMPLETE',
    }

    def __init__(self):
        """Mixin for reporting status states and strings"""
        pass

    @property
    def status_code(self) -> int:
        """returns a status code representative of the state of the Action"""
        return 0

    @property
    def status_string(self) -> str:
        """string meaning of status code"""
        return self.ACTION_STATES[self.status_code]

    @property
    def status(self) -> str:
        """status string for the Action"""
        return f'{self.status_code}: {self.status_string}'

    def as_dict(self) -> dict:
        """represents the status as a dictionary"""
        return {
            'status_code': self.status_code,
            'status': self.status_string,
        }



import threading
import time
import unittest
from enum import IntEnum
from hein_control.states import (ComponentState, OperationalState, ensure_component_online, ensure_component_inactive,
                                 BadComponentState, GenericActionState)


class CustomActiveState(IntEnum):
    IDLE = 0
    ACTIVE = 1
    HYPERACTIVE = 2
    JUMP_JUMP = 3
    BOOGIE_WOOGIE = 4
    # can you tell I have a two year old?


class TestComponent(ComponentState):
    action_states = CustomActiveState

    @ensure_component_online
    def only_run_online(self):
        return

    @ensure_component_inactive
    def only_run_inactive(self):
        return


class TestComponentStates(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.lock = threading.Lock()
        cls.component_one = TestComponent('test_component1', 1)
        cls.component_two = ComponentState('test_component2', 1)
        cls.component_three = ComponentState('test_component3', 1)

    def set_and_check_state(self, component: ComponentState, attribute: str, state: IntEnum):
        """generic tester for action and operational state"""
        # test by value
        setattr(component, attribute, state.value)
        self.assertIs(getattr(component, attribute), state)
        # test by name
        setattr(component, attribute, state.name)
        self.assertIs(getattr(component, attribute), state)
        # test name capitalization support
        setattr(component, attribute, state.name.lower())
        self.assertIs(getattr(component, attribute), state)
        # test direct state setting
        setattr(component, attribute, state)
        self.assertIs(getattr(component, attribute), state)

    def set_and_check_operational_state(self, component: ComponentState, state: OperationalState):
        """tests the setting of an operational state by state, value, and name"""
        # test property setting
        self.set_and_check_state(component, 'current_operational_state', state)
        # test method updating
        component.update_operational_state(state)
        self.assertIs(component.current_operational_state, state)

    def set_and_check_action_state(self, component: ComponentState, state: OperationalState):
        """tests the setting of an active state by state, value, and name"""
        # test property setting
        self.set_and_check_state(component, 'current_action_state', state)
        # test method updating
        component.update_action_state(state)
        self.assertIs(component.current_action_state, state)

    def test_component_state_setting(self):
        """tests setting and retrieval of component states"""
        with self.lock:
            for state in OperationalState:
                self.set_and_check_operational_state(self.component_one, state)

    def test_online_only(self):
        """tests the online check decorator functionality"""
        with self.lock:
            try:
                # check that the method runs
                self.component_one.only_run_online()
                # set to offline state
                self.component_one.current_operational_state = 0
                self.assertRaises(BadComponentState, self.component_one.only_run_online)
                # set to error state
                self.component_one.current_operational_state = -1
                self.assertRaises(BadComponentState, self.component_one.only_run_online)
            finally:
                self.component_one.current_operational_state = OperationalState.ONLINE

    def test_component_activity_setting(self):
        """tests the setting of the component activity"""
        with self.lock:
            for state in GenericActionState:
                self.set_and_check_action_state(self.component_two, state)

    def test_custom_activity_setting(self):
        """tests components with custom activity support"""
        with self.lock:
            for state in CustomActiveState:
                self.set_and_check_action_state(self.component_one, state)

    def test_inactive_only(self):
        """tests the inactive-only decorator functionality"""
        with self.lock:
            try:
                # set to active state
                for state in CustomActiveState:
                    self.component_one.current_action_state = state
                    if state == 0:
                        # check that the method runs
                        self.component_one.only_run_inactive()
                    else:
                        self.assertRaises(BadComponentState, self.component_one.only_run_inactive)
            finally:
                self.component_one.current_action_state = 0

    def update_component_state(self, state):
        """sets component one to an error state after 1 second"""
        time.sleep(1)
        self.component_one.current_operational_state = state

    def get_component_state_thread(self, state: int = -1):
        """retrieves a thread which sets the first component to the specified state"""
        return threading.Thread(
            target=self.update_component_state,
            daemon=True,
            args=[state],
        )

    def test_self_waiting(self):
        """tests waiting on a component where the error flag is switched mid-wait"""
        with self.lock:
            try:
                # ensure state monitor actually sleeps
                self.component_one.components_state_monitor_sleep(0.1)
                # tests that error raises
                first_thread = self.get_component_state_thread()
                first_thread.start()
                self.assertRaises(
                    BadComponentState,
                    self.component_one.components_state_monitor_sleep,
                    2
                )

            finally:
                self.component_one.current_operational_state = 1

    def test_adjacent_waiting(self):
        """tests a wait where the error flag is flipped on another component"""
        with self.lock:
            try:
                # ensure state monitor actually sleeps
                self.component_two.components_state_monitor_sleep(0.1)
                # perform wait on another component and ensure that bad state propagates to an error
                second_thread = self.get_component_state_thread()
                second_thread.start()
                self.assertRaises(
                    BadComponentState,
                    self.component_two.components_state_monitor_sleep,
                    2
                )

            finally:
                self.component_one.current_operational_state = 1

    def test_solo_monitor(self):
        """tests a wait where a single component is monitored for an error state (ignores errors from other components"""
        with self.lock:
            try:
                # perform wait on an third component, specifically only monitoring that component
                #   component 1 is still in an error state for this test
                third_thread = self.get_component_state_thread()
                third_thread.start()
                self.component_three.components_state_monitor_sleep(
                    1.5,
                    'test_component3'
                )
            finally:
                self.component_one.current_operational_state = 1

    def test_wait_for_operational(self):
        """tests the functionality for waiting for a component to be operational"""
        with self.lock:
            try:
                # set to offline
                self.component_one.current_operational_state = 0
                thread = self.get_component_state_thread(1)
                thread.start()
                self.component_one.wait_for_component_operational_state(1)
            finally:
                self.component_one.current_operational_state = 1



"""
Tools for executing a series of actions where the next action is determined by the previous one (linked by next_action).
Automation sequences may include conditional next actions governed by DecisionActions.
"""

import logging
import warnings
from typing import Union, Callable


from ..action.basic import Action
from ..action.configured import ConfiguredAction, ConfiguredActionList, TrackedActionList, ThreadedTrackedAction
from ..action.operators import run_automation

logger = logging.getLogger(__name__)


def build_automation_sequence(*actions: Union[str, Callable, Action, ConfiguredAction]) -> ConfiguredActionList:
    """
    Builds an automation sequence of actions based on the order provided.

    :param actions: actions to link
    :return: built automation sequence of actions
    """
    out = ConfiguredActionList()
    for ind, action in enumerate(actions):
        # if not a step, create a step instance
        if isinstance(action, ConfiguredAction) is False:
            action = ConfiguredAction(action)
        # if there is a previous step, set next step attribute
        if ind > 0:
            out[-1].next_action = action
        out.append(action)
    return out


class ConfiguredAutomation(ConfiguredAction):
    UUID_PREFIX = 'cauto-'

    def __init__(self,
                 first_action: Union[Callable, Action, str, ConfiguredAction],
                 automation_name: str = None,
                 **kwargs,
                 ):
        """
        A class for configuring an automation run

        :param first_action: first action to run
        :param automation_name: name for the automation
        :param kwargs:
        """
        ConfiguredAction.__init__(
            self,
            action=run_automation,
            configuration_name=automation_name,
            **kwargs
        )
        self._first_action: ConfiguredAction = None
        self.first_action = first_action
        # todo add check that each step in the sequence has a next step defined

    @property
    def first_action(self) -> ConfiguredAction:
        """the first action in the automation run"""
        return self._first_action

    @first_action.setter
    def first_action(self, value: ConfiguredAction):
        if isinstance(value, ConfiguredAction) is False:
            value = ConfiguredAction(value)
        self._first_action = value

    def get_tracked_from_config(self, **kwargs) -> "Automation":
        """
        Gets an automation instance from the configured automation

        :param kwargs: additional kwargs that differ from the instance configuration
        :return: automation instance
        """
        return Automation(configuration=self, **kwargs)


class Automation(ThreadedTrackedAction):
    UUID_PREFIX = 'auto-'

    def __init__(self,
                 configuration: Union[str, ConfiguredAutomation],
                 **kwargs,
                 ):
        """
            An Automation is used to run a sequence of Steps. For an Automation, you need to specify what the first step to
        run is.

        An Automation instance is callable. When called, it will run the first Step, get the next_step from
        that first step, and loop through running and retrieving the next Step to run, until there are no more Steps to run.

        The IfStep enables decision making in an Automation

        The Automation class is also important, because all the arguments for all the Steps must be passed into the
        Automation on instantiation, and these arguments may potentially be updated by the Steps. All these arguments for
        the CustomSteps must be passed into an Automation sequence as a dictionary, where the keys are the argument names
        and the values are the initial values for these arguments. These arguments get stored as self.kwargs in the
        Automation instance, and self.kwargs gets used as the arguments to run the CustomSteps. If the CustomStep run
        function has a return value, it MUST be a dictionary, where the keys are argument names (that must correspond to
        a key in the self.kwargs dictionary of the Automation instance), and the values are the new values for the
        arguments; and in this way the arguments for the Automation instance can be updated.

        In short for using the Automation and Steps classes:
        * An Automation instantiation MUST include a dictionary of arguments and initial values to be used by the Steps
            in the Automation, unless no variables need to be passed between/used by multiple Steps
        * The parameter names for all the functions of CustomSteps must be identical to the keys in the kwargs
            dictionary that get passed into an Automation instance on instantiation
        * The return value of any CustomStep must be a dictionary, with the keys being identical to the keys in the
            kwargs dictionary of the Automation instance, and the values are the updated values
        * All these functions for the CustomSteps must also have a **kwargs parameter, to catch arguments that get passed
            through that are not actually required by the functions.
        * An Automation instance must know what the first step to run is
        * Steps need to be 'chained' to each other

        :param Step, first_action: first step in the automation to run
        :param dict, kwargs: all the entry keys need to be identical to the parameter names for the callable actions
            for all the Steps in the automation
        """
        if isinstance(configuration, ConfiguredAutomation) is False:
            configuration = ConfiguredAutomation.class_instance_by_id(configuration)
        self._configuration: ConfiguredAutomation = None  # update type hint
        ThreadedTrackedAction.__init__(
            self,
            configuration=configuration,
            **kwargs,
        )

    def __repr__(self):
        return f'{self.__class__.__name__} first action: {self.configuration}'

    def __call__(self, **kwargs):
        logger.info(f'beginning execution of automation')
        super(Automation, self).__call__(
            self.configuration.first_action,
            **kwargs
        )

    @property
    def configuration(self) -> ConfiguredAutomation:
        """the configured sequence associated with the instance"""
        return self._configuration

    @configuration.setter
    def configuration(self, value: ConfiguredAutomation):
        if isinstance(value, ConfiguredAutomation) is False:
            raise TypeError(f'a ConfiguredSequence must be used as a configuration for {self.__class__.__name__}')
        self._configuration = value
        self.update_args(*self._configuration.args)
        self.update_kwargs(*self._configuration.kwargs)

    @property
    def first_step(self) -> ConfiguredAction:
        """The first step in the Automation run"""
        warnings.warn(  # v6.0.0
            'first_step has been deprecated, access and set first_action instead',
            DeprecationWarning,
            stacklevel=2,
        )
        return self.configuration.first_action

    @property
    def executed_actions(self) -> TrackedActionList:
        """executed actions of the automation run"""
        if self.complete is True:
            return self.action_return

    def run_automation(self,
                       **kwargs,
                       ):
        """
        Run the first step in the automation, and get the next step that results from it. then loop, running the next
        step after that and so on, until the end of the automation has been reached. Will get called when an
        Automation instance is called.

        :param dict, kwargs: dictionary of arguments to run the Steps in the automation. Nothing should need to be
            passed in here for this because the kwargs should have been passed in when the Automation instance was
            instantiated, but if they were not or need to be updated, they can be passed here
        :return:
        """
        warnings.warn(  # v6.0.0
            'the run_automation step as been deprecated, call the class instance directly',
            DeprecationWarning,
            stacklevel=2,
        )
        self.__call__(**kwargs)

    def copy_configuration(self) -> "Automation":
        """
        Copies the configuration of a Automation instance.

        :return: Automation instance with identical configuration
        """
        return self.__class__(
            self.configuration,
            **self.kwargs,
        )


